
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

static int init= 0;

static double eye[3];
static double target[3];
static double right[3];
static double up[3];
static double focal;
static double nearPoint;
static double farPoint;

static double savedEye[3];
static double savedTarget[3];
static double savedRight[3];
static double savedUp[3];
static double savedNear;
static double savedFar;

void renormalize()
{
    double n;
    double v[3];
    v[0]= eye[0]-target[0];
    v[1]= eye[1]-target[1];
    v[2]= eye[2]-target[2];
    n= (v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
    if(n<1e-15)
    {
	resetView();
	return;
    }

    n= system::sqrt(n);
    v[0]/= n;
    v[1]/= n;
    v[2]/= n;

    n= (right[0]*right[0] + right[1]*right[1] + right[2]*right[2]);
    n= system::sqrt(n);
    right[0]/= n;
    right[1]/= n;
    right[2]/= n;

    up[0]= v[1]*right[2] - v[2]*right[1];
    up[1]= v[2]*right[0] - v[0]*right[2];
    up[2]= v[0]*right[1] - v[1]*right[0];

    n= (up[0]*up[0] + up[1]*up[1] + up[2]*up[2]);
    n= system::sqrt(n);
    up[0]/= n;
    up[1]/= n;
    up[2]/= n;

    right[0]= up[1]*v[2] - up[2]*v[1];
    right[1]= up[2]*v[0] - up[0]*v[2];
    right[2]= up[0]*v[1] - up[1]*v[0];
}

void resetView()
{
    eye[0]= 0.0;
    eye[1]= 0.0;
    eye[2]= 5.0;

    right[0]= 1.0;
    right[1]= 0.0;
    right[2]= 0.0;

    up[0]= 0.0;
    up[1]= 1.0;
    up[2]= 0.0;

    target[0]= 0.0;
    target[1]= 0.0;
    target[2]= 0.0;

    nearPoint= 1e-3;
    farPoint= 1e+3;
    focal= 45.0;

    trash();
}

void setView()
{
    if(init==0)
    {
	init= 1;
	resetView();
    }

    double proj[16];
    double dz= farPoint-nearPoint;
    double angle= (focal/2.0)*system::pi()/180.0;
    double s= system::sin(angle);
    double ratio= windowWidth()/(double)windowHeight();
    if(dz==0.0 || s==0.0 || ratio==0.0) return;

    double cot= system::cos(angle)/s;
    proj[ 0]= cot/ratio;
    proj[ 1]= 0.0;
    proj[ 2]= 0.0;
    proj[ 3]= 0.0;

    proj[ 4]= 0.0;
    proj[ 5]= cot;
    proj[ 6]= 0.0;
    proj[ 7]= 0.0;

    proj[ 8]= 0.0;
    proj[ 9]= 0.0;
    proj[10]= -(farPoint+nearPoint)/dz;
    proj[11]= -2.0*nearPoint*farPoint/dz;

    proj[12]= 0.0;
    proj[13]= 0.0;
    proj[14]= -1.0;
    proj[15]= 0.0;

    trsp44(proj);
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixd(proj);

    proj[ 0]= right[0];
    proj[ 1]= right[1];
    proj[ 2]= right[2];
    proj[ 3]= -(right[0]*eye[0]+right[1]*eye[1]+right[2]*eye[2]);

    proj[ 4]= up[0];
    proj[ 5]= up[1];
    proj[ 6]= up[2];
    proj[ 7]= -(up[0]*eye[0]+up[1]*eye[1]+up[2]*eye[2]);

    double v[3];
    v[0]= eye[0]-target[0];
    v[1]= eye[1]-target[1];
    v[2]= eye[2]-target[2];

    double n= (v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
    if(n)
    {
	n= system::sqrt(n);
	v[0]/= n;
	v[1]/= n;
	v[2]/= n;
    }

    proj[ 8]= v[0];
    proj[ 9]= v[1];
    proj[10]= v[2];
    proj[11]= -(v[0]*eye[0]+v[1]*eye[1]+v[2]*eye[2]);

    proj[12]= 0.0;
    proj[13]= 0.0;
    proj[14]= 0.0;
    proj[15]= 1.0;

    trsp44(proj);
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixd(proj);

    renormalize();
}

void saveView()
{
    savedEye[0]= eye[0];
    savedEye[1]= eye[1];
    savedEye[2]= eye[2];

    savedTarget[0]= target[0];
    savedTarget[1]= target[1];
    savedTarget[2]= target[2];

    savedRight[0]= right[0];
    savedRight[1]= right[1];
    savedRight[2]= right[2];

    savedUp[0]= up[0];
    savedUp[1]= up[1];
    savedUp[2]= up[2];

    savedNear= nearPoint;
    savedFar= farPoint;
}

void zoom(
double f
)
{
    double v[3];
    v[0]= savedEye[0]-savedTarget[0];
    v[1]= savedEye[1]-savedTarget[1];
    v[2]= savedEye[2]-savedTarget[2];

    double n= v[0]*v[0]+v[1]*v[1]+v[2]*v[2];
    if(n<1e-10 && f<1.0)	return;
    if(f>1e+10 && f>1.0)	return;
    if(f==1.0)		return;

    eye[0]= target[0] + f*v[0];
    eye[1]= target[1] + f*v[1];
    eye[2]= target[2] + f*v[2];

    nearPoint= savedNear*f;
    farPoint= savedFar*f;

    trash();
}

void tumble(
double xAngle,
double yAngle
)
{
    double m[16];
    rot44(m,savedTarget,savedRight,xAngle);
    apply44To3(eye,m,savedEye);
    apply44To3v(up,m,savedUp);

    double u[3];
    u[0]= 0.0;
    u[1]= 1.0;
    u[2]= 0.0;
    rot44(m,savedTarget,u,yAngle);
    apply44To3(eye,m,eye);
    apply44To3v(right,m,savedRight);
    apply44To3v(up,m,up);

    trash();
}

void pan(
double dx,
double dy
)
{
    double n;
    double v[3];
    v[0]= eye[0] - target[0];
    v[1]= eye[1] - target[1];
    v[2]= eye[2] - target[2];
    n= v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
    n= system::sqrt(n);
    n*= 2.0*system::tan((focal/2.0)*system::pi()/180.0);

    dx*= n;
    dy*= n;

    double tx= -dx*right[0] + dy*up[0];
    double ty= -dx*right[1] + dy*up[1];
    double tz= -dx*right[2] + dy*up[2];

    eye[0]= savedEye[0] + tx;
    eye[1]= savedEye[1] + ty;
    eye[2]= savedEye[2] + tz;

    target[0]= savedTarget[0] + tx;
    target[1]= savedTarget[1] + ty;
    target[2]= savedTarget[2] + tz;

    trash();
}

void paintCamera()
{
    if(getInteractionState())
    {
	double ratio= windowWidth()/(double)windowHeight();
	glMatrixMode(GL_MODELVIEW); glPushMatrix();glLoadIdentity();
	glMatrixMode(GL_PROJECTION); glPushMatrix();glLoadIdentity();
	    glColor3d(1.0,0.0,0.0);
	    glBegin(GL_LINES);
		glVertex3d(-0.1/ratio,0.0,0.0);
		glVertex3d(+0.1/ratio,0.0,0.0);
		glVertex3d(0.0,-0.1,0.0);
		glVertex3d(0.0,+0.1,0.0);
	    glEnd();
	glMatrixMode(GL_PROJECTION); glPopMatrix();
	glMatrixMode(GL_MODELVIEW); glPopMatrix();
    }
}

void centerViewOnFix(
double la,
double lo
)
{
    double vn;
    double view[3];
    view[0]= eye[0]-target[0];
    view[1]= eye[1]-target[1];
    view[2]= eye[2]-target[2];
    vn= system::sqrt(view[0]*view[0] + view[1]*view[1] + view[2]*view[2]);

    double r[3];
    double u[3];
    double v[3];
    localFrame(la,lo,r,u,v);

    target[0]= r[0]*(1.0-1e-10);
    target[1]= r[1]*(1.0-1e-10);
    target[2]= r[2]*(1.0-1e-10);

    eye[0]= target[0]+vn*r[0];
    eye[1]= target[1]+vn*r[1];
    eye[2]= target[2]+vn*r[2];

    right[0]= u[0];
    right[1]= u[1];
    right[2]= u[2];

    up[0]= v[0];
    up[1]= v[1];
    up[2]= v[2];

    trash();
}

void getEye(
double *_eye
)
{
    _eye[0]= eye[0]; 
    _eye[1]= eye[1]; 
    _eye[2]= eye[2];
}

void getEyePickLine(
double *_eye,
double *_dir,
double x,
double y
)
{
    x*= 2.0/(double)(windowWidth()-1);
    y*= 2.0/(double)(windowHeight()-1);
    x-= 1.0;
    y-= 1.0;

    double ratio= windowWidth()/(double)windowHeight();
    x*=ratio;

    double v[3];
    v[0]= right[1]*up[2] - right[2]*up[1];
    v[1]= right[2]*up[0] - right[0]*up[2];
    v[2]= right[0]*up[1] - right[1]*up[0];

    _eye[0]= eye[0]; 
    _eye[1]= eye[1]; 
    _eye[2]= eye[2];

    double cot= 1.0/system::tan((focal/2.0)*system::pi()/180.0);
    _dir[0]= x*right[0] + y*up[0] - cot*v[0];
    _dir[1]= x*right[1] + y*up[1] - cot*v[1];
    _dir[2]= x*right[2] + y*up[2] - cot*v[2];

    double n= _dir[0]*_dir[0] + _dir[1]*_dir[1] + _dir[2]*_dir[2];
    n= system::sqrt(n);
    if(n)
    {
	_dir[0]/= n;
	_dir[1]/= n;
	_dir[2]/= n;
    }
}

