
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

void localFrame(
double la,	// Latitude
double lo,	// Longitude
double *r,	// Radial vector
double *u,	// Tangent vector to latitude towards the east
double *v	// Tangent vector to meridians towards the north
)
{
    double rla= la*system::pi()/180.0;
    double cla= system::cos(rla);
    double sla= system::sin(rla);

    double rlo= lo*system::pi()/180.0;
    double clo= system::cos(rlo);
    double slo= system::sin(rlo);

    if(r)
    {
	r[0]= cla*slo;
	r[1]= sla;
	r[2]= cla*clo;
    }

    if(u)
    {
	u[0]= clo;
	u[1]= 0.0;
	u[2]= -slo;
    }

    if(v)
    {
	v[0]= -sla*slo;
	v[1]= cla;
	v[2]= -sla*clo;
    }
}

int parseNMEAAngle(
const uint8_t *angle,
const uint8_t *dir,
double *result
)
{
    const uint8_t *p= angle;
    while(1)
    {
	if(p[0]==0) return -1;
	else if(p[0]=='.') break;
	++p;
    }

    p-=2;
    if(p<angle) return -1;
    int min= (p[0]-'0')*10 + (p[1]-'0');

    int deg= 0;
    const uint8_t *s= angle;
    while(s<p)
    {
	deg*= 10;
	deg+= (s[0]-'0');
	++s;
    }

    int div= 1;
    const uint8_t *e= p+3;
    double fracMin= 0.0;
    while(e[0])
    {
	div*= 10;
	fracMin*=10.0;
	fracMin+= e[0]-'0';
	++e;
    }

    double dMin= ((double)min) + (fracMin/(double)div);
    result[0]= (((double)deg) + dMin/60.0) * ((dir[0]=='S' || dir[0]=='W') ? -1.0 : 1.0);
    return 0;
}

static double la;
static double lo;
static int autoMap= 0;
static double stamp= -1;

double lastLat()	{ return la;	}
double lastLon()	{ return lo;	}
double lastStamp()	{ return stamp;	}

void toggleAutomap()
{
    autoMap^= 1;
    trash();
}

int getAutoMap()
{
    return autoMap;
}

static uint8_t *utc;
static uint8_t *quality;
static uint8_t *altitude;

void newGGA(
const uint8_t *_utc,
const uint8_t *_quality,
const uint8_t *_altitude
)
{
    if(utc) system::free(utc);
    if(quality) system::free(quality);
    if(altitude) system::free(altitude);

    utc= (uint8_t*)system::strdup((const char*)_utc);
    quality= (uint8_t*)system::strdup((const char*)_quality);
    altitude= (uint8_t*)system::strdup((const char*)_altitude);

    trash();
}

void newFix(
const uint8_t *latitude,
const uint8_t *latitudeSign,
const uint8_t *longitude,
const uint8_t *longitudeSign,
const uint8_t *timeStamp
)
{
    double tmpLa;
    double tmpLo;
    if(parseNMEAAngle(latitude,latitudeSign,&tmpLa)<0) return;
    if(parseNMEAAngle(longitude,longitudeSign,&tmpLo)<0) return;
    if(tmpLa!=la || tmpLo!=lo)
    {
	double hr= (timeStamp[0]-'0')*10 + (timeStamp[1]-'0');
	double mm= (timeStamp[2]-'0')*10 + (timeStamp[3]-'0');
	double ss= (timeStamp[4]-'0')*10 + (timeStamp[5]-'0');
	ss+= (timeStamp[6]-'0')/10.0 + (timeStamp[7]-'0')/100.0;
	ss+= mm*60;
	ss+= hr*3600;

	la= tmpLa;
	lo= tmpLo;
	stamp= ss;
	trash();

    }

    if(autoMap)
    {
	centerViewOnFix(la,lo);

	double eye[3];
	double pos[3];
	getEye(eye);
	localFrame(la,lo,pos,0,0);

	pos[0]-= eye[0];
	pos[1]-= eye[1];
	pos[2]-= eye[2];

	pos[0]*= pos[0];
	pos[1]*= pos[1];
	pos[2]*= pos[2];

	double dist= system::sqrt(pos[0] + pos[1] + pos[2]);
	getHTTPMap(0,la,lo,dist);
    }
}

static void unitCircle(
double	frac,
int	addStipple= 0
)
{
    if(frac>0)
    {
	double step= 0.01;
	double tMin= -system::pi()*frac;
	double tMax=  system::pi()*frac;
	glBegin(GL_LINE_STRIP);
	    int done= 0;
	    double theta= tMin;
	    while(done==0)
	    {
		if(theta>tMax)
		{
		    theta= tMax;
		    done= 1;
		}

		double x= system::sin(theta);
		double y= 0.0;
		double z= system::cos(theta);
		glVertex3d(x,y,z);

		theta+= step;
	    }
	glEnd();
	if(addStipple)
	{
	    glColor3ub(0xFF,0x00,0x00);
	    glBegin(GL_LINE_STRIP);
		int done= 0;
		double theta= tMax;
		while(done==0)
		{
		    if(theta>(2.0-frac)*system::pi())
		    {
			theta= (2.0-frac)*system::pi();
			done= 1;
		    }

		    double x= system::sin(theta);
		    double y= 0.0;
		    double z= system::cos(theta);
		    glVertex3d(x,y,z);

		    theta+= step;
		}
	    glEnd();
	}
    }
}

static void laArc(
double	_la,
double	_lo,
double	frac
)
{
    glPushMatrix();
	double rla= _la*system::pi()/180.0;
	double cla= system::cos(rla);
	double sla= system::sin(rla);
	glTranslated(0.0,sla,0.0);
	glScaled(cla,1.0,cla);
	glRotated(_lo,0.0,1.0,0.0);
	unitCircle(frac);
    glPopMatrix();
}

static void loArc(
double	_la,
double	_lo,
double	frac
)
{
    glPushMatrix();
	glRotated(_lo,0.0,1.0,0.0);
	glRotated(-_la,1.0,0.0,0.0);
	glRotated(90,0.0,0.0,1.0);
	unitCircle(frac);
    glPopMatrix();
}

void paintFix(
const	uint8_t	*name,
	double	_la,
	double	_lo,
	double	frac
)
{
    double rla= _la*system::pi()/180.0;
    double cla= system::cos(rla);
    double sla= system::sin(rla);

    double rlo= _lo*system::pi()/180.0;
    double clo= system::cos(rlo);
    double slo= system::sin(rlo);

    double x= cla*slo;
    double y= sla;
    double z= cla*clo;

    double eye[3];
    getEye(eye);
    
    bool doIt= true;
    double n= eye[0]*eye[0] + eye[1]*eye[1] + eye[2]*eye[2];
    if(n>0.9)
    {
	double dot= (eye[0]*x + eye[1]*y + eye[2]*z);
	if(dot<0.0) doIt= false;
    }

    if(doIt==true)
    {
	loArc(_la,_lo,frac);
	laArc(_la,_lo,frac);

	if(name)
	{

	    glRasterPos3d(x,y,z);
	    paintString(name);
	}
    }
}

void paintSquare(
double	la0,
double	lo0,
double	la1,
double	lo1
)
{
    double fracLo= system::fabs(lo1-lo0)/360.0;
    double fracLa= system::fabs(la1-la0)/360.0;
    laArc(la0,(lo0+lo1)/2.0,fracLo);
    laArc(la1,(lo0+lo1)/2.0,fracLo);
    loArc((la0+la1)/2.0,lo0,fracLa);
    loArc((la0+la1)/2.0,lo1,fracLa);
}

void paintGreatcircle(
double	la0,
double	lo0,
double	la1,
double	lo1,
int	addStipple,
double	*r,
double	*a
)
{
    double r0[3];
    double r1[3];
    localFrame(la0,lo0,r0,0,0);
    localFrame(la1,lo1,r1,0,0);

    double cr[3];
    cr[0]= r0[1]*r1[2]-r0[2]*r1[1];
    cr[1]= r0[2]*r1[0]-r0[0]*r1[2];
    cr[2]= r0[0]*r1[1]-r0[1]*r1[0];

    double radius= 6378.136;
    double cosAlpha= r0[0]*r1[0] + r0[1]*r1[1] + r0[2]*r1[2];
    double sinAlpha= system::sqrt(cr[0]*cr[0] + cr[1]*cr[1] + cr[2]*cr[2]);
    double alpha= system::atan2(sinAlpha,cosAlpha);

    if(sinAlpha)
    {
	glPushMatrix();
	    cr[0]/= sinAlpha;
	    cr[1]/= sinAlpha;
	    cr[2]/= sinAlpha;

	    double r2[3];
	    r2[0]= r0[1]*cr[2]-r0[2]*cr[1];
	    r2[1]= r0[2]*cr[0]-r0[0]*cr[2];
	    r2[2]= r0[0]*cr[1]-r0[1]*cr[0];

	    double mat[16];
	    mat[ 0]= r2[0]; mat[ 4]= cr[0]; mat[ 8]= r0[0]; mat[12]= 0.0;
	    mat[ 1]= r2[1]; mat[ 5]= cr[1]; mat[ 9]= r0[1]; mat[13]= 0.0;
	    mat[ 2]= r2[2]; mat[ 6]= cr[2]; mat[10]= r0[2]; mat[14]= 0.0;
	    mat[ 3]=     0; mat[ 7]=     0; mat[11]=     0; mat[15]= 1.0;

	    glMultMatrixd(mat);
	    glRotated(-180.0/system::pi()*alpha/2.0,0.0,1.0,0.0);
	    unitCircle(alpha/(2.0*system::pi()),addStipple);
	glPopMatrix();
    }

    if(r) *r= radius;
    if(a) *a= alpha;
}

void paintSquareInfo(
double	x,
double	y,
double	la0,
double	lo0,
double	la1,
double	lo1
)
{
    double radius;
    double alpha;
    paintGreatcircle(la0,lo0,la1,lo1,1,&radius,&alpha);

    glColor3ub(0xFF,0xFF,0x00);
    glMatrixMode(GL_MODELVIEW); glPushMatrix();glLoadIdentity();
    glMatrixMode(GL_PROJECTION); glPushMatrix();glLoadIdentity();
	glOrtho(
	    0.0, (double)windowWidth(),
	    0.0, (double)windowHeight(),
	    -1000.0, 1000.0
	);

	uint8_t tab[1024];
	system::sprintf((char*)tab,"Selection info: Great circle= %.2f km ",radius*alpha);
	glRasterPos3d(x,y,0.0);
	paintString(tab);

	double box[4];
	stringBBox(tab,box);

	system::sprintf((char*)tab,"(%.2f km)",radius*(2.0*system::pi()-alpha));
	glColor3ub(0xFF,0x00,0x00);
	glRasterPos3d(x+box[2]-box[0],y,0.0);
	paintString(tab);

	system::sprintf((char*)tab,"Lat range=  [%+010.4f   %+010.4f]",la0,la1);
	glColor3ub(0xFF,0xFF,0x00);
	glRasterPos3d(x,y-fontHeight(),0.0);
	paintString(tab);

	system::sprintf((char*)tab,"Lon range= [%+010.4f   %+010.4f]",lo0,lo1);
	glRasterPos3d(x,y-2.0*fontHeight(),0.0);
	paintString(tab);
    glMatrixMode(GL_PROJECTION); glPopMatrix();
    glMatrixMode(GL_MODELVIEW); glPopMatrix();
}

void paintFixCoords(
const	uint8_t	*name,
	double	xPos,
	double	yPos,
	double	_la,
	double	_lo
)
{
    glMatrixMode(GL_MODELVIEW); glPushMatrix();glLoadIdentity();
    glMatrixMode(GL_PROJECTION); glPushMatrix();glLoadIdentity();
	glOrtho(
	    0.0, (double)windowWidth(),
	    0.0, (double)windowHeight(),
	    -1000.0, 1000.0
	);

	uint8_t tab[1024];
	system::sprintf((char*)tab,"%s:",name);
	glRasterPos3d(xPos,yPos,0.0);
	paintString(tab);

	system::sprintf((char*)tab,"Lat=  %+010.4f",_la);
	glRasterPos3d(xPos,yPos-fontHeight(),0.0);
	paintString(tab);

	system::sprintf((char*)tab,"Lon= %+010.4f",_lo);
	glRasterPos3d(xPos,yPos-2.0*fontHeight(),0.0);
	paintString(tab);
    glMatrixMode(GL_PROJECTION); glPopMatrix();
    glMatrixMode(GL_MODELVIEW); glPopMatrix();
}

void paintLastFix()
{
    glColor3ub(0xFF,0x00,0x00);
    paintFix((const uint8_t*)"Current fix",la,lo,1.0/32.0);
    paintFixCoords((const uint8_t*)"Current fix",10.0,37.0,la,lo);
}

void paintAuxInfo()
{
    if(utc==0) return;

    glMatrixMode(GL_MODELVIEW); glPushMatrix();glLoadIdentity();
    glMatrixMode(GL_PROJECTION); glPushMatrix();glLoadIdentity();
	glOrtho(
	    0.0, (double)windowWidth(),
	    0.0, (double)windowHeight(),
	    -1000.0, 1000.0
	);

	char tab[1024];
	glColor3ub(0xFF,0x00,0x00);

	system::sprintf(tab,"GPS signal quality: %s\n",quality);
	glRasterPos3d((double)windowWidth()-160.0,10+2*fontHeight(),0.0);
	paintString((const uint8_t*)tab);

	system::sprintf(tab,"Fix altitude (meters): %s\n",altitude);
	glRasterPos3d((double)windowWidth()-160.0,10+0*fontHeight(),0.0);
	paintString((const uint8_t*)tab);

	system::sprintf(
	    tab,
	    "UTC: %c%ch %c%cm %ss",
	    utc[0],
	    utc[1],
	    utc[2],
	    utc[3],
	    utc+4
	);
	glRasterPos3d((double)windowWidth()-160.0,10+1*fontHeight(),0.0);
	paintString((const uint8_t*)tab);


    glMatrixMode(GL_PROJECTION); glPopMatrix();
    glMatrixMode(GL_MODELVIEW); glPopMatrix();
}
